/*

Copyright (c) 2015 GostCrypt team <gostcrypt_info@gostcrypt.org>

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the Free
software Foundation; either version 2 of the License, or (at your option)
any later version.

*/

#include "EncryptionModeCM.h"
#include "Common/Crypto.h"


namespace GostCrypt
{
	void EncryptionModeGCM::Decrypt(byte* data, uint64 length) const
	{
		DecryptBuffer(data, length);
	}

	void EncryptionModeGCM::Encrypt(byte* data, uint64 length) const
	{
		EncryptBuffer(data, length);
	}

	void EncryptionModeGCM::DecryptBuffer(byte* data, uint64 length)
	{
		if_debug (ValidateState());

		CipherList::const_iterator iSecondaryCipher = SecondaryCiphers.begin();

		for (CipherList::const_iterator iCipher = Ciphers.begin(); iCipher != Ciphers.end(); ++iCipher)
		{
			if ((**iCipher).GetBlockSize() == 8)
				DecryptBufferCM8Bytes (**iCipher, **iSecondaryCipher, data, length, startDataUnitNo, 0);
			else
				DecryptBufferCM (**iCipher, **iSecondaryCipher, data, length, startDataUnitNo, 0);
			++iSecondaryCipher;
		}

		assert (iSecondaryCipher == SecondaryCiphers.end());
	}

	void EncryptionModeGCM::EncryptBuffer(byte* data, uint64 length)
	{
		if_debug (ValidateState());

		CipherList::const_iterator iSecondaryCipher = SecondaryCiphers.begin();

		for (CipherList::const_iterator iCipher = Ciphers.begin(); iCipher != Ciphers.end(); ++iCipher)
		{
			if ((**iCipher).GetBlockSize() == 8)
				EncryptBufferCM8Bytes (**iCipher, **iSecondaryCipher, data, length, startDataUnitNo, 0);
			else
				EncryptBufferCM (**iCipher, **iSecondaryCipher, data, length, startDataUnitNo, 0);
			++iSecondaryCipher;
		}

		assert (iSecondaryCipher == SecondaryCiphers.end());
	}

/*	void EncryptionModeGCM::DecryptBufferCM8Bytes(const Cipher &cipher, const Cipher &secondaryCipher, byte *buffer, uint64 length, uint64 startDataUnitNo, unsigned int startCipherBlockNo) const
	{

	}*/
	
	void EncryptionModeGCM::DecryptBufferCM(const Cipher &cipher, const Cipher &secondaryCipher, byte *buffer, uint64 length, uint64 startDataUnitNo, unsigned int startCipherBlockNo) const
	{
		unsigned long int H[2] = {0x0000000000000000, 0x0000000000000000};
		unsigned long int IV;
		byte xor_ks[MAX_EXPANDED_KEY];
		byte xor_ks_2[MAX_EXPANDED_KEY];
		uint64 counter[2];
		uint64 blockcount;


		IV = startDataUnitNo;
		cipher.CopyCipherKey((byte*) xor_ks);
		
		//Définir un H, valeur initiale de l'algo
		Cipher.EncryptWithKS((byte*) H, xor_ks);

		//Définition de l'IV
		//Calcul de GHASH(H, NULL, IV) // Considérons que le IV est startDataUnitNo chiffré avec le secondaryCipher
		secondaryCipher.CopyCipherKey((byte*) xor_ks_2);
		secondaryCipher.EncryptWithKS((byte*)IV, xor_ks_2);
		GHASH(H, NULL, IV, 0, 8, &counter); //H, A, IV, lenA, lenIV, valeur de retour

		blockcount = length / BYTES_PER_GCM_BLOCK;

		//Le traitement est différent pour le dernier bloc
		while(blockcount > 0)
		{

			Cipher.DecryptWithKS((byte*) counter, xor_ks);
			buffer[0] ^= counter[0];
			buffer[1] ^= counter[1];

			//On vient de chiffrer un bloc
			blockcount--;
			counter++;

		}
	}

/*	void EncryptionModeGCM::EncryptBufferCM8Bytes(const Cipher &cipher, const Cipher &secondaryCipher, byte *buffer, uint64 length, uint64 startDataUnitNo, unsigned int startCipherBlockNo) const
	{

	}*/

	void EncryptionModeGCM::EncryptBufferCM(const Cipher &cipher, const Cipher &secondaryCipher, byte *buffer, uint64 length, uint64 startDataUnitNo, unsigned int startCipherBlockNo) const
	{
		unsigned long int H[2] = {0x0000000000000000, 0x0000000000000000};
		unsigned long int IV;
		byte xor_ks[MAX_EXPANDED_KEY];
		byte xor_ks_2[MAX_EXPANDED_KEY];
		uint64 counter[2];
		uint64 blockcount;


		IV = startDataUnitNo;
		cipher.CopyCipherKey((byte*) xor_ks);
		
		//Définir un H, valeur initiale de l'algo
		Cipher.EncryptWithKS((byte*) H, xor_ks);

		//Définition de l'IV
		//Calcul de GHASH(H, NULL, IV) // Considérons que le IV est startDataUnitNo chiffré avec le secondaryCipher
		secondaryCipher.CopyCipherKey((byte*) xor_ks_2);
		secondaryCipher.EncryptWithKS((byte*)IV, xor_ks_2);
		GHASH(H, NULL, IV, 0, 8, &counter); //H, A, IV, lenA, lenIV, valeur de retour

		blockcount = length / BYTES_PER_GCM_BLOCK;

		//Le traitement est différent pour le dernier bloc
		while(blockcount > 0)
		{

			Cipher.EncryptWithKS((byte*) counter, xor_ks);
			buffer[0] ^= counter[0];
			buffer[1] ^= counter[1];

			//On vient de chiffrer un bloc
			blockcount--;
			counter++;

		}

	}

	void EncryptionModeGCM::initIV(unsigned long long *IV, unsigned char *Y)
	{
   		unsigned long long Z[2];
    	short i;
    	Z[0] = IV[0]; Z[1] = IV[1];
    	Z[0] <<= 32; Z[0] |= (Z[1] >> 32) & 0xFFFFFFFF;
    	Z[1] <<= 32;
    	Z[1] |= 0x00000001;
    	IV[0] = Z[0];
    	IV[1] = Z[1];
    
    	for (i=0; i < 16; i++)
    	{
        	if (i < 8)
            	Y[i] = IV[0] >> (56 - 8*i);
        	else
            	Y[i] = IV[1] >> (120 - 8*i); // 56 - 8*(i-8)
    	}
	}

	void EncryptionModeGCM::GHASH()
	{}
}















