/*

Copyright (c) 2015 GostCrypt team <gostcrypt_info@gostcrypt.org>

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the Free
software Foundation; either version 2 of the License, or (at your option)
any later version.
*/

#ifndef GST_HEADER_Volume_EncryptionModeGCM
#define GST_HEADER_Volume_EncryptionModeGCM

#include "Platform/Platform.h"
#include "EncryptionMode.h"

namespace GostCrypt
{
	class EncryptionModeGCM : public EncryptionMode
	{
	public:
		EncryptionModeGCM () { } //Constructeur
		virtual ~EncryptionModeGCM () { } //Destructeur

		virtual void Decrypt (byte *data, uint64 length) const;
		virtual void DecryptSectorsCurrentThread (byte *data, uint64 sectorIndex, uint64 sectorCount, size_t sectorSize) const;
		virtual void Encrypt (byte *data, uint64 length) const;
		virtual void EncryptSectorsCurrentThread (byte *data, uint64 sectorIndex, uint64 sectorCount, size_t sectorSize) const;
		virtual const SecureBuffer &GetKey () const { return SecondaryKey; }
		virtual size_t GetKeySize () const;
		virtual wstring GetName () const { return L"GCM"; };
		virtual shared_ptr <EncryptionMode> GetNew () const { return shared_ptr <EncryptionMode> (new EncryptionModeGCM); }
		virtual void SetCiphers (const CipherList &ciphers);
		virtual void SetKey (const ConstBufferPtr &key);

		
	protected:
		void DecryptBuffer(byte* data, uint64 length) const;
		void EncryptBuffer(byte* data, uint64 length) const;
		void initIV(unsigned long long *IV, unsigned char *Y);

		SecureBuffer SecondaryKey;
		CipherList SecondaryCiphers;
		
	private:
		EncryptionModeGCM (const EncryptionModeGCM &);
		EncryptionModeGCM &operator= (const EncryptionModeGCM &);
	};
}

#endif //GST_HEADER_Volume_EncryptionModeGCM